#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 32
void NES_mapper32::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  patch = 0;

  // for Major League
  if(parent_NES->sum_16k_rom() == 0x1d1084)
  {
    set_mirroring(0,0,0,0);
    patch = 1;
  }
  // for Ai Sensei no Oshiete - Watashi no Hoshi
  if(parent_NES->sum_16k_rom() == 0x140108)
  {
    set_CPU_banks(30,31,30,31);
  }

  regs[0] = 0;
}

void NES_mapper32::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF000)
  {
    case 0x8000:
      {
        if(regs[0] & 0x02)
        {
          set_CPU_bank6(data);
        }
        else
        {
          set_CPU_bank4(data);
        }
      }
      break;

    case 0x9000:
      {
        if(data & 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        regs[0] = data;
      }
      break;

    case 0xA000:
      {
        set_CPU_bank5(data);
      }
      break;
  }

  switch(addr & 0xF007)
  {
    case 0xB000:
      {
        set_PPU_bank0(data);
      }
      break;

    case 0xB001:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0xB002:
      {
        set_PPU_bank2(data);
      }
      break;

    case 0xB003:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0xB004:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0xB005:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0xB006:
      {
        if ((patch == 1) && (data & 0x40))
        {
          set_mirroring(0,0,0,1);
        }
        set_PPU_bank6(data);
      }
      break;

    case 0xB007:
      {
        if ((patch == 1) && (data & 0x40))
        {
          set_mirroring(0,0,0,0);
        }
        set_PPU_bank7(data);
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 65
void NES_mapper65::Reset()
{
  patch = 0;

  // for Kaiketsu Yanchamaru 3
  if(parent_NES->sum_16k_rom() == 0x189be4)
  {
    patch = 1;
  }

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_enabled = 0;
  irq_counter = 0;
}

void NES_mapper65::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x9000:
      {
        if(!patch)
        {
          if(data & 0x40)
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
        }
      }
      break;

    case 0x9001:
      {
        if(patch)
        {
          if(data & 0x80)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
        }
      }
      break;

    case 0x9003:
      {
        if(!patch)
        {
          irq_enabled = data & 0x80;
        }
      }
      break;

    case 0x9004:
      {
        if(!patch)
        {
          irq_counter = irq_latch;
        }
      }
      break;

    case 0x9005:
      {
        if(patch)
        {
          irq_counter = (uint8)(data << 1);
          irq_enabled = data;
        }
        else
        {
          irq_latch = (irq_latch & 0x00FF) | ((uint32)data << 8);
        }
      }
      break;

    case 0x9006:
      {
        if(patch)
        {
          irq_enabled = 1;
        }
        else
        {
          irq_latch = (irq_latch & 0xFF00) | data;
        }
      }
      break;

    case 0xB000:
      {
        set_PPU_bank0(data);
      }
      break;

    case 0xB001:
      {
        set_PPU_bank1(data);
      }
      break;

    case 0xB002:
      {
        set_PPU_bank2(data);
      }
      break;

    case 0xB003:
      {
        set_PPU_bank3(data);
      }
      break;

    case 0xB004:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0xB005:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0xB006:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0xB007:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0xA000:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0xC000:
      {
        set_CPU_bank6(data);
      }
      break;
  }
}

void NES_mapper65::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(patch)
    {
      if(irq_counter == 0)
      {
        parent_NES->cpu->DoIRQ();
        irq_enabled = 0;
      }
      else
      {
        irq_counter--;
      }
    }
    else
    {
      if(irq_counter <= 113)
      {
        parent_NES->cpu->DoIRQ();
        irq_enabled = 0;
        irq_counter = 0xFFFF;
      }
      else
      {
        irq_counter -= 113;
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 77
void NES_mapper77::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);

  // for Napoleon Senki
  parent_NES->ppu->vram_write_protect = 0;
}

void NES_mapper77::MemoryWrite(uint32 addr, uint8 data)
{
  uint8 prg_bank = data & 0x07;
  uint8 chr_bank = (data & 0xF0) >> 4;

  set_CPU_bank4(prg_bank*4+0);
  set_CPU_bank5(prg_bank*4+1);
  set_CPU_bank6(prg_bank*4+2);
  set_CPU_bank7(prg_bank*4+3);

  set_PPU_bank0(chr_bank*2+0);
  set_PPU_bank1(chr_bank*2+1);
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 97
void NES_mapper97::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(num_8k_ROM_banks-2,num_8k_ROM_banks-1,0,1);

  // set PPU bank pointers ?
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper97::MemoryWrite(uint32 addr, uint8 data)
{
  if(addr < 0xC000)
  {
    uint8 prg_bank = data & 0x0F;

    set_CPU_bank6(prg_bank*2+0);
    set_CPU_bank7(prg_bank*2+1);

    if((data & 0x80) == 0)
    {
      set_mirroring(NES_PPU::MIRROR_HORIZ);
    }
    else
    {
      set_mirroring(NES_PPU::MIRROR_VERT);
    }
  }
}
/////////////////////////////////////////////////////////////////////

#endif
