#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 67
void NES_mapper67::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  set_PPU_bank0(0);
  set_PPU_bank1(1);
  set_PPU_bank2(2);
  set_PPU_bank3(3);
  set_PPU_bank4(num_8k_ROM_banks-4);
  set_PPU_bank5(num_8k_ROM_banks-3);
  set_PPU_bank6(num_8k_ROM_banks-2);
  set_PPU_bank7(num_8k_ROM_banks-1);

  irq_enabled = 0;
  irq_counter = 0;
  irq_latch = 0;
}

void NES_mapper67::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF800)
  {
    case 0x8800:
      {
        set_PPU_bank0(data*2);
        set_PPU_bank1(data*2+1);
      }
      break;

    case 0x9800:
      {
        set_PPU_bank2(data*2);
        set_PPU_bank3(data*2+1);
      }
      break;

    case 0xA800:
      {
        set_PPU_bank4(data*2);
        set_PPU_bank5(data*2+1);
      }
      break;

    case 0xB800:
      {
        set_PPU_bank6(data*2);
        set_PPU_bank7(data*2+1);
      }
      break;

    case 0xC800:
      {
        irq_counter = irq_latch;
        irq_latch = data;
      }
      break;

    case 0xD800:
      {
        irq_enabled = data & 0x10;
      }
      break;

    case 0xE800:
      {
        data &= 0x03;
        if(data == 0)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else if(data == 1)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else if(data == 2)
        {
          set_mirroring(0,0,0,0);
        }
        else
        {
          set_mirroring(1,1,1,1);
        }
      }
      break;

    case 0xF800:
      {
        set_CPU_bank4(data*2);
        set_CPU_bank5(data*2+1);
      }
      break;
  }
}

void NES_mapper67::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(--irq_counter == 0xF6)
        {
          irq_counter = irq_latch;
          parent_NES->cpu->DoIRQ();
        }
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 68
void NES_mapper68::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  regs[0] = 0;
  regs[1] = 0;
  regs[2] = 0;
  regs[3] = 0;
}

void NES_mapper68::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xF000)
  {
    case 0x8000:
      {
        set_PPU_bank0(data*2+0);
        set_PPU_bank1(data*2+1);
      }
      break;

    case 0x9000:
      {
        set_PPU_bank2(data*2+0);
        set_PPU_bank3(data*2+1);
      }
      break;

    case 0xA000:
      {
        set_PPU_bank4(data*2+0);
        set_PPU_bank5(data*2+1);
      }
      break;

    case 0xB000:
      {
        set_PPU_bank6(data*2+0);
        set_PPU_bank7(data*2+1);
      }
      break;

    case 0xC000:
      {
        regs[2] = data;
        SyncMirror();
      }
      break;

    case 0xD000:
      {
        regs[3] = data;
        SyncMirror();
      }
      break;

    case 0xE000:
      {
        regs[0] = (data & 0x10) >> 4;
        regs[1] = data & 0x03;
        SyncMirror();
      }
      break;

    case 0xF000:
      {
        set_CPU_bank4(data*2);
        set_CPU_bank5(data*2+1);
      }
      break;
  }
}

void NES_mapper68::SyncMirror()
{
  if(regs[0])
  {
    if (regs[1] == 0)
    {
      set_PPU_bank8(regs[2] + 0x80);   // + 0x20000
      set_PPU_bank9(regs[3] + 0x80);
      set_PPU_bank10(regs[2] + 0x80);
      set_PPU_bank11(regs[3] + 0x80);
    }
    else if (regs[1] == 1)
    {
      set_PPU_bank8(regs[2] + 0x80);
      set_PPU_bank9(regs[2] + 0x80);
      set_PPU_bank10(regs[3] + 0x80);
      set_PPU_bank11(regs[3] + 0x80);
    }
    else if (regs[1] == 2)
    {
      set_PPU_bank8(regs[2] + 0x80);
      set_PPU_bank9(regs[2] + 0x80);
      set_PPU_bank10(regs[2] + 0x80);
      set_PPU_bank11(regs[2] + 0x80);
    }
    else if (regs[1] == 3)
    {
      set_PPU_bank8(regs[3] + 0x80);
      set_PPU_bank9(regs[3] + 0x80);
      set_PPU_bank10(regs[3] + 0x80);
      set_PPU_bank11(regs[3] + 0x80);
    }
  }
  else
  {
    if (regs[1] == 0)
    {
      set_mirroring(NES_PPU::MIRROR_VERT);
    }
    else if (regs[1] == 1)
    {
      set_mirroring(NES_PPU::MIRROR_HORIZ);
    }
    else if (regs[1] == 2)
    {
      set_mirroring(0,0,0,0);
    }
    else if (regs[1] == 3)
    {
      set_mirroring(1,1,1,1);
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 69
void NES_mapper69::Reset()
{
  patch = 0;

  // for Dynamite Batman
  if(parent_NES->sum_16k_rom() == 0x20b0f1)
  {
    patch = 1;
  }

  // Init ExSound
  parent_NES->apu->SelectExSound(32);

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_enabled = 0;
  irq_counter = 0;
}

void NES_mapper69::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr & 0xE000)
  {
    case 0x8000:
      {
        regs[0] = data;
      }
      break;

    case 0xA000:
      {
        switch(regs[0] & 0x0F)
        {
          case 0x00:
            {
              set_PPU_bank0(data);
            }
            break;

          case 0x01:
            {
              set_PPU_bank1(data);
            }
            break;

          case 0x02:
            {
              set_PPU_bank2(data);
            }
            break;

          case 0x03:
            {
              set_PPU_bank3(data);
            }
            break;

          case 0x04:
            {
              set_PPU_bank4(data);
            }
            break;

          case 0x05:
            {
              set_PPU_bank5(data);
            }
            break;

          case 0x06:
            {
              set_PPU_bank6(data);
            }
            break;

          case 0x07:
            {
              set_PPU_bank7(data);
            }
            break;

          case 0x08:
            {
              if(!patch)
              {
                if(!(data & 0x40))
                {
                  set_CPU_bank3(data);
                }
              }
            }
            break;

          case 0x09:
            {
              set_CPU_bank4(data);
            }
            break;

          case 0x0A:
            {
              set_CPU_bank5(data);
            }
            break;

          case 0x0B:
            {
              set_CPU_bank6(data);
            }
            break;

          case 0x0C:
            {
              data &= 0x03;
              if(data == 0)
              {
                set_mirroring(NES_PPU::MIRROR_VERT);
              }
              else if(data == 1)
              {
                set_mirroring(NES_PPU::MIRROR_HORIZ);
              }
              else if(data == 2)
              {
                set_mirroring(0,0,0,0);
              }
              else
              {
                set_mirroring(1,1,1,1);
              }
            }
            break;

          case 0x0D:
            {
              irq_enabled = data;
            }
            break;

          case 0x0E:
            {
              irq_counter = (irq_counter & 0xFF00) | data;
            }
            break;

          case 0x0F:
            {
              irq_counter = (irq_counter & 0x00FF) | (data << 8);
            }
            break;
        }
      }
      break;

    case 0xC000:
    case 0xE000:
      {
        parent_NES->apu->ExWrite(addr, data);
      }
      break;
  }
}

void NES_mapper69::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if(irq_counter <= 113)
    {
      parent_NES->cpu->DoIRQ();
      irq_counter = 0;
    }
    else
    {
      irq_counter -= 113;
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 89
void NES_mapper89::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper89::MemoryWrite(uint32 addr, uint8 data)
{
  if((addr & 0xFF00) == 0xC000)
  {
    uint8 prg_bank = (data & 0x70) >> 4;
    uint8 chr_bank = ((data & 0x80) >> 4) | (data & 0x07);

    set_CPU_bank4(prg_bank*2+0);
    set_CPU_bank5(prg_bank*2+1);

    set_PPU_bank0(chr_bank*8+0);
    set_PPU_bank1(chr_bank*8+1);
    set_PPU_bank2(chr_bank*8+2);
    set_PPU_bank3(chr_bank*8+3);
    set_PPU_bank4(chr_bank*8+4);
    set_PPU_bank5(chr_bank*8+5);
    set_PPU_bank6(chr_bank*8+6);
    set_PPU_bank7(chr_bank*8+7);

    if (data & 0x08)
    {
      set_mirroring(1,1,1,1);
    }
    else
    {
      set_mirroring(0,0,0,0);
    }
  }
}

/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 93
void NES_mapper93::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);
  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);
}

void NES_mapper93::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(addr == 0x6000)
  {
    set_CPU_bank4(data*2+0);
    set_CPU_bank5(data*2+1);
  }
}

/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 122
void NES_mapper122::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,2,3);
}

void NES_mapper122::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  if(addr = 0x6000)
  {
    uint8 chr_bank0 = data & 0x07;
    uint8 chr_bank1 = (data & 0x70) >> 4;

    set_PPU_bank0(chr_bank0*4+0);
    set_PPU_bank1(chr_bank0*4+1);
    set_PPU_bank2(chr_bank0*4+2);
    set_PPU_bank3(chr_bank0*4+3);
    set_PPU_bank4(chr_bank1*4+0);
    set_PPU_bank5(chr_bank1*4+1);
    set_PPU_bank6(chr_bank1*4+2);
    set_PPU_bank7(chr_bank1*4+3);
  }
}

/////////////////////////////////////////////////////////////////////

#endif
