#ifdef _NES_MAPPER_CPP_

/////////////////////////////////////////////////////////////////////
// Mapper 33
void NES_mapper33::Reset()
{
  patch = 0;

  if(parent_NES->sum_16k_rom() == 0x138606 || // Akira
     parent_NES->sum_16k_rom() == 0x1203d3 || // Bakushou!! Jinsei Gekijou
     parent_NES->sum_16k_rom() == 0x11dbc5 || // Don Doko Don
     parent_NES->sum_16k_rom() == 0x18de80 || // Golf Ko Open
     parent_NES->sum_16k_rom() == 0x2ae94a || // Operation Wolf
     parent_NES->sum_16k_rom() == 0x1aeab1 || // Power Blazer
     parent_NES->sum_16k_rom() == 0x20a4d5)   // Takeshi no Sengoku Fuuunji
  {
    patch = 1;
  }

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  irq_enabled = 0;
  irq_counter = 0;
}

void NES_mapper33::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        if(patch) // old #33 games
        {
          if(data & 0x40)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
          set_CPU_bank4(data & 0x1F);
        }
        else
        {
          set_CPU_bank4(data);
        }
      }
      break;

    case 0x8001:
      {
        if(patch) // old #33 games
        {
          set_CPU_bank5(data & 0x1F);
        }
        else
        {
          set_CPU_bank5(data);
        }
      }
      break;

    case 0x8002:
      {
        set_PPU_bank0(data*2+0);
        set_PPU_bank1(data*2+1);
      }
      break;

    case 0x8003:
      {
        set_PPU_bank2(data*2+0);
        set_PPU_bank3(data*2+1);
      }
      break;

    case 0xA000:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0xA001:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0xA002:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0xA003:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0xC000:
      {
        irq_counter = data;
      }
      break;

    case 0xC001:
    case 0xC002:
    case 0xE001:
    case 0xE002:
      {
        irq_enabled = data;
      }
      break;

    case 0xE000:
      {
        if (data & 0x40)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
      }
      break;
  }
}

void NES_mapper33::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if (scanline >= 0 && scanline <= 239)
    {
      if(irq_counter == 0xFF)
      {
        parent_NES->cpu->DoIRQ();
        irq_counter = 0;
        irq_enabled = 0;
      }
      else
      {
        irq_counter++;
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 48
void NES_mapper48::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  set_PPU_banks(0,1,2,3,4,5,6,7);

  regs[0] = 0;
  irq_enabled = 0;
  irq_counter = 0;
}

void NES_mapper48::MemoryWrite(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x8000:
      {
        if(!regs[0])
        {
          if(data & 0x40)
          {
            set_mirroring(NES_PPU::MIRROR_HORIZ);
          }
          else
          {
            set_mirroring(NES_PPU::MIRROR_VERT);
          }
        }
        set_CPU_bank4(data);
      }
      break;

    case 0x8001:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x8002:
      {
        set_PPU_bank0(data*2+0);
        set_PPU_bank1(data*2+1);
      }
      break;

    case 0x8003:
      {
        set_PPU_bank2(data*2+0);
        set_PPU_bank3(data*2+1);
      }
      break;

    case 0xA000:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0xA001:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0xA002:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0xA003:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0xC000:
      {
        irq_counter = data;
      }
      break;

    case 0xC001:
      {
        irq_enabled = data & 0x01;
      }
      break;

    case 0xE000:
      {
        if (data & 0x40)
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        regs[0] = 1;
      }
      break;
  }
}

void NES_mapper48::HSync(uint32 scanline)
{
  if(irq_enabled)
  {
    if((scanline >= 0) && (scanline <= 239))
    {
      if(parent_NES->ppu->spr_enabled() || parent_NES->ppu->bg_enabled())
      {
        if(irq_counter == 0xFF)
        {
          parent_NES->cpu->DoIRQ();
          irq_enabled = 0;
        }
        else
        {
          irq_counter++;
        }
      }
    }
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 80
void NES_mapper80::Reset()
{
  patch = 0;

  if(num_8k_ROM_banks == 32)
  {
    // for Hudoumyouou Den
    patch = 1;
  }

  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }
}

void NES_mapper80::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  switch(addr)
  {
    case 0x7EF0:
      {
        set_PPU_bank0(data & 0x7F);
        set_PPU_bank1((data & 0x7F)+1);

        if(patch)
        {
          if (data & 0x80)
          {
            parent_NES->ppu->set_PPU_VRAM(0x8, 1);
            parent_NES->ppu->set_PPU_VRAM(0x9, 1);
          }
          else
          {
            parent_NES->ppu->set_PPU_VRAM(0x8, 0);
            parent_NES->ppu->set_PPU_VRAM(0x9, 0);
          }
        }
      }
      break;

    case 0x7EF1:
      {
        set_PPU_bank2(data & 0x7F);
        set_PPU_bank3((data & 0x7F)+1);

        if(patch)
        {
          if (data & 0x80)
          {
            parent_NES->ppu->set_PPU_VRAM(0xA, 1);
            parent_NES->ppu->set_PPU_VRAM(0xB, 1);
          }
          else
          {
            parent_NES->ppu->set_PPU_VRAM(0xA, 0);
            parent_NES->ppu->set_PPU_VRAM(0xB, 0);
          }
        }
      }
      break;

    case 0x7EF2:
      {
        set_PPU_bank4(data);
      }
      break;

    case 0x7EF3:
      {
        set_PPU_bank5(data);
      }
      break;

    case 0x7EF4:
      {
        set_PPU_bank6(data);
      }
      break;

    case 0x7EF5:
      {
        set_PPU_bank7(data);
      }
      break;

    case 0x7EF6:
      {
        if(data & 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
      }
      break;

    case 0x7EFA:
    case 0x7EFB:
      {
        set_CPU_bank4(data);
      }
      break;

    case 0x7EFC:
    case 0x7EFD:
      {
        set_CPU_bank5(data);
      }
      break;

    case 0x7EFE:
    case 0x7EFF:
      {
        set_CPU_bank6(data);
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Mapper 82
void NES_mapper82::Reset()
{
  // set CPU bank pointers
  set_CPU_banks(0,1,num_8k_ROM_banks-2,num_8k_ROM_banks-1);

  // set PPU bank pointers
  if(num_1k_VROM_banks)
  {
    set_PPU_banks(0,1,2,3,4,5,6,7);
  }

  // set Mirroring
  set_mirroring(NES_PPU::MIRROR_VERT);

  regs[0] = 0;
}

void NES_mapper82::MemoryWriteSaveRAM(uint32 addr, uint8 data)
{
  switch (addr)
  {
    case 0x7EF0:
      /* Switch 2k VROM at $0000 or $1000 */
      {
        if(regs[0])
        {
          set_PPU_bank4((data & 0xFE)+0);
          set_PPU_bank5((data & 0xFE)+1);
        }
        else
        {
          set_PPU_bank0((data & 0xFE)+0);
          set_PPU_bank1((data & 0xFE)+1);
        }
      }
      break;

    case 0x7EF1:
      {
        if(regs[0])
        {
          set_PPU_bank6((data & 0xFE)+0);
          set_PPU_bank7((data & 0xFE)+1);
        }
        else
        {
          set_PPU_bank2((data & 0xFE)+0);
          set_PPU_bank3((data & 0xFE)+1);
        }
      }
      break;

    case 0x7EF2:
      {
        if(!regs[0])
        {
          set_PPU_bank4(data);
        }
        else
        {
          set_PPU_bank0(data);
        }
      }
      break;

    case 0x7EF3:
      {
        if(!regs[0])
        {
          set_PPU_bank5(data);
        }
        else
        {
          set_PPU_bank1(data);
        }
      }
      break;

    case 0x7EF4:
      {
        if(!regs[0])
        {
          set_PPU_bank6(data);
        }
        else
        {
          set_PPU_bank2(data);
        }
      }
      break;

    case 0x7EF5:
      {
        if(!regs[0])
        {
          set_PPU_bank7(data);
        }
        else
        {
          set_PPU_bank3(data);
        }
      }
      break;

    case 0x7EF6:
      {
        regs[0] = data & 0x02;
        if(data & 0x01)
        {
          set_mirroring(NES_PPU::MIRROR_VERT);
        }
        else
        {
          set_mirroring(NES_PPU::MIRROR_HORIZ);
        }
      }
      break;

    case 0x7EFA:
      {
        set_CPU_bank4(data >> 2);
      }
      break;

    case 0x7EFB:
      {
        set_CPU_bank5(data >> 2);
      }
      break;

    case 0x7EFC:
      {
        set_CPU_bank6(data >> 2);
      }
      break;
  }
}
/////////////////////////////////////////////////////////////////////

#endif
